/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  localcache.c -- cache code if we are doing our own caching. This is
  experimental.

TODO Config file options.
  ***************************************/
#include <sys/stat.h>
#include <fcntl.h>
#include "common.h"
#include "configutil.h"
#include "cache.h"
#include "control.h"

extern int cachemgr_init(char *cd, int cs);

char *cached_filename(const unsigned char *uri);
int file_cached(const char *uri, const char *mdtm, int size, int offset);

static char *cfile;
static int fd, cmgrfd;
static enum{CACHED, CACHING, NONE} cache_status=NONE;

static char *cachedir;

int l_geninit(char **cacheopts, char *chroot)
{
	int cs=0;
	char *p, **pp;

	for(pp=cacheopts; *pp!=NULL; pp++){
		p=strtok(*pp, " \t=");
		if(!strcasecmp(p, "Dir")) cachedir=strdup(strtok(NULL, " \t=\n"));
		else if(!strcasecmp(p, "CacheSize")) cs=atoi(strtok(NULL, " \t\n"));
		else return(-1);
	}
	strip_filename(chroot, cachedir);
	if(strlen(cachedir)==0 || cachedir[strlen(cachedir)-1]!='/')
		strcat(cachedir,"/");
	cmgrfd=cachemgr_init(cachedir, cs * 1024 * 1024);
	return(0);
}

/* ------------------------------------------------------------- ** 
** If we have the file return a fd to it. Otherwise return -1.
** ------------------------------------------------------------- */
int l_retr_start(const char *host, const char *file, const char *mdtm, int size, int offset)
{
	char uri[512];

	if(*mdtm==0){
		debug("Server didn't like MDTM. Can't cache :(\n");
		cache_status=NONE;
		return(-1);
	}

	strcpy(uri, "ftp://");
	strncat(uri, host, 511-strlen(uri));
	uri[511]=0;
	strncat(uri, file, 511-strlen(uri));
	uri[511]=0;

	if(!file_cached(uri, mdtm, size, offset)) {
		debug("Cache miss. Forwarding RETR to server\n");
		return(-1);
	}
	debug("Cache hit. Returning file\n");

	send_message(150,"Starting transfer");

	return(fd);
}

/* ------------------------------------------------------------- **
** Called whenever we select on incoming data from server.
** Write data to cache.
** ------------------------------------------------------------- */
void l_inc_data(char *inc, int *len)
{
	if(cache_status==CACHING) write(fd, inc, *len);
}

void l_retr_end(void)
{
	if(cache_status==CACHED) {
		send_message(226, "Transfer complete");
	}
	if(cache_status==CACHING){
		close(fd);
		write(cmgrfd, "N",1);
		write(cmgrfd, cfile, strlen(cfile));
		write(cmgrfd, "\n",1);
	}
	cache_status=NONE;
}

/* ------------------------------------------------------------- ** 
** Return TRUE if we have an up to date copy of the file. 
** Also set global variable fd to the file if we have it, or to where
** we should write the file if we don't.
**
** Each cache file has a header of "NNNN MDTM SIZE TYPE URI\n" where
** NNNN is the number of bytes in the header not including the "NNNN ".
** MDTM and SIZE are the returns from those functions, and TYPE is 0 
** for ascii, and 1 for binary.
**
** REST handling is fairly simplistic. If offset>0 and we have the
** portion of the file which is not requested then we seek to offset
** in it and overwrite with what we retrieve to complete our cached
** copy. Otherwise we give up. 
** ------------------------------------------------------------- */
int file_cached(const char *uri, const char *mdtm, int size, int offset)
{
	char buf[MAX_LINE_LEN*2];
	struct stat stats;
	int type=1; /*FIXME*/

	cfile=cached_filename(uri);

	if(stat(cfile, &stats)!=-1){
		int i, fsize, ftype;
		char *fmdtm, *furi;

		fd = offset ? open(cfile, O_RDWR) : open(cfile, O_RDONLY);

		read(fd, buf, 5);
		i=atoi(buf);
		if( i < MAX_LINE_LEN*2){
			if(read(fd, buf, i)==i && offset<=stats.st_size-i-5 ){

				fmdtm=strtok(buf, " ");
				fsize=atoi(strtok(NULL, " "));
				ftype=atoi(strtok(NULL, " "));
				furi=strtok(NULL, " \r\n");
				
				if(!strcmp(fmdtm, mdtm) && 
				   fsize==size &&
				   ftype==type &&
				   !strcmp(furi, uri)){
					lseek(fd, offset, SEEK_CUR);
					if(stats.st_size-i-5 == size){
						cache_status=CACHED;
						write(cmgrfd, "A",1);
						write(cmgrfd, cfile, strlen(cfile));
						write(cmgrfd, "\n",1);
						return(TRUE);
					} else { /*We'll read from offset onwards*/
						cache_status=CACHING;
						return(FALSE);
					}
				}
			}
		}
		debug("Cached file is not valid. Deleting.\n");
		close(fd);
		unlink(cfile);
	}

	/* We've been asked for a portion of file and we don't have it. 
	 * We could try and be clever and save this portion so that if
	 * someone later wants the whole thing we don't need it all,
	 * but it probably isn't worth it. Give up.*/
	if(offset>0) {
		cache_status=NONE;
		return(FALSE);
	}

	fd=creat(cfile, S_IRUSR | S_IWUSR);
	if(fd==-1){
		write_log(ERROR, "Creating cache file %s failed." 
			  "Stopping caching", cfile);
		cache_status=NONE;
		return(FALSE);
	}
	sprintf(buf, "     %s %d %d %s\n", mdtm, 
		size, type, uri);
	sprintf(buf, "%d", strlen(buf)-5);
	buf[strlen(buf)]=' ';
	write(fd, buf, strlen(buf));
	cache_status=CACHING;
	return(FALSE);
}

/* ------------------------------------------------------------- **
** The following is an MD5sum algorithm brutally hacked from Colin
** Plumb's public domain code. 
** ------------------------------------------------------------- */
#define UWORD32 u_int32_t
void MD5Transform(UWORD32 buf[4], UWORD32 const in[16]);

#ifdef WORDS_BIGENDIAN
void byteSwap(UWORD32 *buf, unsigned words)
{
	md5byte *p = (md5byte *)buf;

	do {
		*buf++ = (UWORD32)((unsigned)p[3] << 8 | p[2]) << 16 |
			((unsigned)p[1] << 8 | p[0]);
		p += 4;
	} while (--words);
}
#else
#define byteSwap(buf,words)
#endif

/* ------------------------------------------------------------- **
** Hash uri and return the filename (with absolute path) that this
** file should be cached in. 
** ------------------------------------------------------------- */
char *cached_filename(const unsigned char *uri)
{
	static char filename[MAX_LINE_LEN];
	UWORD32 buf[4];
	UWORD32 bytes[2];
	UWORD32 in[16];
	unsigned char *p;
	int len, i;

	strcpy(filename, cachedir);

	buf[0] = 0x67452301;
	buf[1] = 0xefcdab89;
	buf[2] = 0x98badcfe;
	buf[3] = 0x10325476;

	bytes[0] = 0;
	bytes[1] = 0;

	len = strlen(uri);
	bytes[0] = len;

	/* Process data in 64-byte chunks */
	while (len >= 64) {
		memcpy(in, uri, 64);
		byteSwap(in, 16);
		MD5Transform(buf, in);
		uri += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */
	memcpy(in, uri, len);

	p = (unsigned char *)in + len;

	/* Set the first char of padding to 0x80.  There is always room. */
	*p++ = 0x80;

	/* Bytes of padding needed to make 56 bytes (-8..55) */
	len = 56 - 1 - len;

	if (len < 0) {	/* Padding forces an extra block */
		memset(p, 0, len + 8);
		byteSwap(in, 16);
		MD5Transform(buf, in);
		p = (unsigned char *)in;
		len = 56;
	}
	memset(p, 0, len);
	byteSwap(in, 14);

	/* Append length in bits and transform */
	in[14] = bytes[0] << 3;
	in[15] = bytes[1] << 3 | bytes[0] >> 29;
	MD5Transform(buf, in);

	byteSwap(buf, 4);

	p=(unsigned char *)buf;
	sprintf(filename+strlen(filename), "%02x/", (*p++) % 0x0F);
	for (i = 0; i < 15; ++i)
                sprintf(filename+strlen(filename), "%02x", *p++);

	return(filename);
}

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f,w,x,y,z,in,s) \
	 (w += f(x,y,z) + in, w = (w<<s | w>>(32-s)) + x)

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
void MD5Transform(UWORD32 buf[4], UWORD32 const in[16])
{
	register UWORD32 a, b, c, d;

	a = buf[0];
	b = buf[1];
	c = buf[2];
	d = buf[3];

	MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
	MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
	MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
	MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
	MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
	MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
	MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
	MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
	MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
	MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
	MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
	MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
	MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
	MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
	MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
	MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

	MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
	MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
	MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
	MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
	MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
	MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
	MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
	MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
	MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
	MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
	MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
	MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
	MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
	MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
	MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
	MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

	MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
	MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
	MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
	MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
	MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
	MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
	MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
	MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
	MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
	MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
	MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
	MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
	MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
	MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
	MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
	MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

	MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
	MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
	MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
	MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
	MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
	MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
	MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
	MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
	MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
	MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
	MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
	MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
	MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
	MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
	MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
	MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);

	buf[0] += a;
	buf[1] += b;
	buf[2] += c;
	buf[3] += d;
}
